"""Tests around the graylog reactive script."""
import os
import socket
import sys
import tempfile
import unittest
from unittest import mock

from charms.layer.graylog.constants import CERT_KEY_PATH, CERT_PATH
from charms.layer.graylog.snap_change import ChannelChangeStatus, _is_channel_valid
from charms.reactive import is_state, set_state

# some dep layers only exists in the built charm; mock them out before
# the graylog imports since those depend on post-build charms.* layers
leader_mock = mock.Mock()
sys.modules["charms.leadership"] = leader_mock
snap_mock = mock.Mock()
sys.modules["charms.layer.snap"] = snap_mock
tls_client_mock = mock.Mock()
sys.modules["charms.layer.tls_client"] = tls_client_mock

from files import check_graylog_health  # noqa: E402

from reactive.graylog import (  # noqa: E402
    GraylogApi,
    _check_input_exists,
    application_dashboard_relation_changed,
    configure_graylog,
    get_total_mem,
    install_graylog,
    refresh_graylog,
    set_conf,
    set_jvm_heap_size,
    tls_request_certificate,
    update_config,
    upgrade_charm,
    validate_jvm_heap_size,
)


class TestInstallUpdateUpgrade(unittest.TestCase):
    """Test that install / update / upgrade clears our configured flag."""

    @mock.patch("reactive.graylog.snap")
    @mock.patch("reactive.graylog.hookenv.config")
    def test_install(self, mock_config, mock_snap):  # noqa: D102
        class Config(dict):
            def __init__(self, *args, **kw):
                super(Config, self).__init__(*args, **kw)

        mock_config.return_value = Config({"channel": "2/stable"})
        install_graylog()
        self.assertTrue(mock_snap.install.called)
        self.assertFalse(is_state("graylog.configured"))

    def test_update(self):  # noqa: D102
        set_state("graylog.configured")
        update_config()
        self.assertFalse(is_state("graylog.configured"))

    def test_upgrade(self):  # noqa: D102
        set_state("graylog.configured")
        upgrade_charm()
        self.assertFalse(is_state("graylog.configured"))


class TestRefresh(unittest.TestCase):
    """Test the snap refresh handler."""

    @mock.patch("charms.layer.graylog.snap_change.snap")
    @mock.patch("reactive.graylog.hookenv.config")
    def test_no_channel(self, mock_config, mock_snap):
        """Ensure a no-op if there is no 'channel' config."""
        mock_config.return_value = ""
        mock_snap.get_installed_version.return_value = "2.foo"
        set_state("graylog.configured")
        refresh_graylog()
        self.assertFalse(mock_snap.refresh.called)
        self.assertTrue(is_state("graylog.configured"))

    # test_change and test_no_change are getting a bit cumbersome due to
    # reworks/refactorings; we may want to consider dropping these 2 and instead
    # relying on functional coverage.

    @mock.patch("charms.layer.graylog.snap_change.unitdata")  # Same as below
    @mock.patch("reactive.graylog.unitdata")  # Avoid unintended local state change
    @mock.patch("charms.layer.graylog.snap_change.os")
    @mock.patch("charms.layer.graylog.snap_change.snap")
    @mock.patch("reactive.graylog.snap")
    @mock.patch("reactive.graylog.hookenv.config")
    def test_no_change(
        self, mock_config, mock_snap, mock_snap2, mock_os, mock_ud, mock_ud2
    ):
        """Ensure no reconfiguration if the snap version does not change."""
        mock_config.return_value = "2/stable"
        mock_snap.get_installed_version.side_effect = (
            mock_snap2.get_installed_version.side_effect
        ) = ["2.1", "2.1"]
        mock_snap.get_installed_channel.return_value = (
            mock_snap2.get_installed_channel.return_value
        ) = "2/stable"
        mock_os.path.exists.return_value = True
        mock_ud.kv.return_value = mock_kv = mock.MagicMock()
        # For update_status, mock that no channel_block_type has been set
        mock_kv.get.return_value = None
        set_state("graylog.configured")
        refresh_graylog()
        self.assertFalse(mock_snap.refresh.called)
        self.assertTrue(is_state("graylog.configured"))

    @mock.patch("charms.layer.graylog.snap_change.unitdata")  # Same as below
    @mock.patch("reactive.graylog.unitdata")  # Avoid unintended local state changes
    @mock.patch("charms.layer.graylog.snap_change.data_changed")
    @mock.patch("charms.layer.graylog.snap_change.shutil.copy2")
    @mock.patch("charms.layer.graylog.snap_change.os")
    @mock.patch("charms.layer.graylog.snap_change.host")
    @mock.patch("charms.layer.graylog.snap_change.snap")
    @mock.patch("reactive.graylog.snap")
    @mock.patch("reactive.graylog.hookenv.config")
    def test_change(
        self,
        mock_config,
        mock_snap,
        mock_snap2,
        mock_host,
        mock_os,
        mock_shutil,
        mock_data_changed,
        mock_ud,
        mock_ud2,
    ):
        """Ensure calls and states when the snap has changed."""
        mock_config.return_value = "2/stable"
        mock_snap.get_installed_version.side_effect = (
            mock_snap2.get_installed_version.side_effect
        ) = ["2.1", "2.2"]
        mock_snap.get_installed_channel.return_value = (
            mock_snap2.get_installed_channel.return_value
        ) = "2/candidate"
        mock_os.path.exists.return_value = True
        mock_shutil.return_value = "foo"
        mock_ud.kv.return_value = mock_kv = mock.MagicMock()
        # For update_status, mock that no channel_block_type has been set
        mock_kv.get.return_value = None
        set_state("graylog.configured")
        refresh_graylog()
        self.assertTrue(mock_host.service_stop.called)
        self.assertTrue(mock_snap2.refresh.called)
        self.assertEqual(mock_shutil.call_count, 2)
        self.assertEqual(mock_data_changed.call_count, 2)
        self.assertFalse(is_state("graylog.configured"))


class TestConfig(unittest.TestCase):
    """Test the configure_graylog handler."""

    @mock.patch("reactive.graylog.hookenv")
    @mock.patch("reactive.graylog.os")
    def test_no_conf_file(self, mock_os, mock_hookenv):
        """Ensure a status message if there is no config file present."""
        mock_os.path.isfile.return_value = False
        mock_os.path.getsize.return_value = 1
        configure_graylog()
        self.assertTrue(mock_hookenv.status_set.called)

    @mock.patch("reactive.graylog.get_channel_change_status")
    @mock.patch("reactive.graylog.is_master")
    @mock.patch("reactive.graylog.set_jvm_heap_size")
    @mock.patch("reactive.graylog.validate_api_uri")
    @mock.patch("charms.layer.graylog.utils.is_v2")
    @mock.patch("reactive.graylog.is_v2")
    @mock.patch("reactive.graylog.set_conf")
    @mock.patch("reactive.graylog.hookenv")
    @mock.patch("reactive.graylog.os")
    @mock.patch("reactive.graylog.unitdata")
    def test_v2_config(
        self,
        mock_ud,
        mock_os,
        mock_hookenv,
        mock_set_conf,
        mock_v2,
        mock_v2_2,
        mock_validate_uri,
        mock_jvm_heap,
        mock_is_master,
        mock_gccs,
    ):
        """Ensure calls and states with various config values for v2."""

        class Config(dict):
            def __init__(self, *args, **kw):
                super(Config, self).__init__(*args, **kw)

        mock_ud.kv.return_value = {
            "admin_password": "password",
            "password_secret": "secret",
        }
        mock_os.path.isfile.return_value = True
        mock_os.path.getsize.return_value = 1
        mock_hookenv.config.return_value = Config(
            {
                "jvm_heap_size": "1G",
                "web_listen_uri": None,
                "rest_transport_uri": "foo/api",
                "web_endpoint_uri": "bar/api",
                "email_enabled": False,
            }
        )
        mock_v2.return_value = mock_v2_2.return_value = True
        mock_is_master.return_value = True
        mock_gccs.return_value = ChannelChangeStatus.NO_CHANGE
        configure_graylog()
        self.assertEqual(mock_validate_uri.call_count, 2)
        self.assertTrue(mock_jvm_heap.called)
        self.assertTrue(mock_hookenv.application_version_set.called)
        self.assertTrue(mock_hookenv.open_port.called)
        self.assertTrue(mock_is_master.called)

    @mock.patch("reactive.graylog.get_channel_change_status")
    @mock.patch("reactive.graylog.is_master")
    @mock.patch("reactive.graylog.set_jvm_heap_size")
    @mock.patch("reactive.graylog.validate_api_uri")
    @mock.patch("charms.layer.graylog.utils.is_v2")
    @mock.patch("reactive.graylog.is_v2")
    @mock.patch("reactive.graylog.set_conf")
    @mock.patch("reactive.graylog.hookenv")
    @mock.patch("reactive.graylog.os")
    @mock.patch("reactive.graylog.unitdata")
    def test_v3_config(
        self,
        mock_ud,
        mock_os,
        mock_hookenv,
        mock_set_conf,
        mock_v2,
        mock_v2_2,
        mock_validate_uri,
        mock_jvm_heap,
        mock_is_master,
        mock_gccs,
    ):
        """Ensure calls and states with various config values for v3."""

        class Config(dict):
            def __init__(self, *args, **kw):
                super(Config, self).__init__(*args, **kw)

        mock_ud.kv.return_value = {
            "admin_password": "password",
            "password_secret": "secret",
        }
        mock_os.path.isfile.return_value = True
        mock_os.path.getsize.return_value = 1
        mock_hookenv.config.return_value = Config(
            {
                "jvm_heap_size": "1G",
                "web_listen_uri": None,
                "rest_transport_uri": "foo/api",
                "web_endpoint_uri": "bar/api",
                "email_enabled": False,
            }
        )
        mock_v2.return_value = mock_v2_2.return_value = False
        mock_is_master.return_value = True
        mock_gccs.return_value = ChannelChangeStatus.NO_CHANGE
        configure_graylog()
        self.assertEqual(mock_validate_uri.call_count, 2)
        self.assertTrue(mock_jvm_heap.called)
        self.assertTrue(mock_hookenv.application_version_set.called)
        self.assertTrue(mock_hookenv.open_port.called)
        self.assertTrue(mock_is_master.called)


initial_conf = u"""#key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
key4 = value5"""

update1_conf = u"""#key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
key4 = valueUpdated"""

commented1_conf = u"""key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
key4 = value5"""

commented2_conf = u"""key1 = value1
key2 = value2
key2=value3
key3 = value4
key4 = value5"""

duplicate1_conf = u"""#key1 = value1
key2 = value6
#key2=value3
### key3   =  value 4
key4 = value5"""

removed_conf = u"""#key1 = value1
key2 = value2
key2=value3
### key3   =  value 4
#key4 = value5"""

log_input_types = {
    "org.graylog.plugins.beats.BeatsInput": "Beats",
    "org.graylog2.inputs.raw.udp.RawUDPInput": "Raw/Plaintext UDP",
}

log_inputs = [
    {
        "title": "Beats Input",
        "global": "true",
        "name": "Beats",
        "created_at": "2017-08-17T10:19:10.634Z",
        "type": "org.graylog.plugins.beats.BeatsInput",
        "creator_user_id": "admin",
        "attributes": {"bind_address": "0.0.0.0", "port": 11044},
        "id": "5994f3aab0d37f23c38adbda",
    }
]


class TestSetConf(unittest.TestCase):
    """Verify Juju config updates."""

    def setUp(self):  # noqa: D102
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        temp_file.write(initial_conf.encode())
        temp_file.close()
        self.conf_file = temp_file.name

    def tearDown(self):  # noqa: D102
        os.remove(self.conf_file)

    def conf_equals(self, want):  # noqa: D102
        with open(self.conf_file, "r") as conf:
            got = conf.read()

            if got == want:
                return True
            print("{}\n != \n{}".format(got, want))

        return False

    def test_update(self):  # noqa: D102
        set_conf("key4", "value5", self.conf_file)
        self.assertTrue(self.conf_equals(initial_conf))

        set_conf("key4", "valueUpdated", self.conf_file)
        self.assertTrue(self.conf_equals(update1_conf))

    def test_set_commented(self):  # noqa: D102
        set_conf("key1", "value1", self.conf_file)
        self.assertTrue(self.conf_equals(commented1_conf))

        set_conf("key3", "value4", self.conf_file)
        self.assertTrue(self.conf_equals(commented2_conf))

    def test_remove_key(self):  # noqa: D102
        set_conf("key4", "", self.conf_file)
        self.assertTrue(self.conf_equals(removed_conf))

    def test_key_not_string(self):  # noqa: D102
        with self.assertRaises(AssertionError):
            set_conf(None, "value", self.conf_file)

    def test_value_not_string(self):  # noqa: D102
        with self.assertRaises(AssertionError):
            set_conf("key1", False, self.conf_file)

    def test_value_none(self):  # noqa: D102
        with self.assertRaises(AssertionError):
            set_conf("key1", None, self.conf_file)

    def test_handle_duplicates(self):  # noqa: D102
        set_conf("key2", "value6", self.conf_file)
        self.assertTrue(self.conf_equals(duplicate1_conf))

    def test_no_changed(self):  # noqa: D102
        self.assertFalse(set_conf("key2", "value2", self.conf_file))

    def test_changed(self):  # noqa: D102
        self.assertTrue(set_conf("key2", "some-new-value", self.conf_file))

    def test_input_type_exists(self):  # noqa: D102
        g = GraylogApi("", "", "")
        g.input_types = log_input_types
        self.assertTrue(g.map_log_input_type("beats"))

    def test_input_type_raw(self):  # noqa: D102
        g = GraylogApi("", "", "")
        g.input_types = log_input_types
        self.assertTrue(g.map_log_input_type("raw udp"))

    def test_input_type_invalid(self):  # noqa: D102
        g = GraylogApi("", "", "")
        g.input_types = log_input_types
        self.assertFalse(g.map_log_input_type("some-weird-type"))

    def test_input_exists(self):  # noqa: D102
        new_input = {
            "title": "Beats Input",
            "type": "org.graylog.plugins.beats.BeatsInput",
            "configuration": {"bind_address": "0.0.0.0", "port": 11044},
        }
        (changed, input_id) = _check_input_exists(log_inputs, new_input)
        self.assertTrue(input_id == "5994f3aab0d37f23c38adbda" and changed is False)

    def test_input_changed(self):  # noqa: D102
        new_input = {
            "title": "Beats Input",
            "type": "org.graylog.plugins.beats.BeatsInput",
            "configuration": {"bind_address": "0.0.0.0", "port": 12222},
        }
        (changed, input_id) = _check_input_exists(log_inputs, new_input)
        self.assertTrue(changed is True)

    def test_input_new(self):  # noqa: D102
        new_input = {
            "title": "Beats Input 2",
            "type": "org.graylog.plugins.beats.BeatsInput",
            "configuration": {"bind_address": "0.0.0.0", "port": 11033},
        }
        (changed, input_id) = _check_input_exists(log_inputs, new_input)
        self.assertTrue(input_id is None)


class TestHealthCheck(unittest.TestCase):
    """Verify Graylog cluster health states."""

    @mock.patch("charms.layer.graylog.api.GraylogApi")
    def test_journal_state(self, mock_api):  # noqa: D102
        ucommitted_tests = [
            ["test_ok", 0, check_graylog_health.STATE_OK],
            ["test_warning", 1, check_graylog_health.STATE_WARNING],
            ["test_critical", 100, check_graylog_health.STATE_CRITICAL],
        ]
        args = mock.MagicMock(uncommitted_warning=0, uncommitted_critical=100)

        for name, ucommitted, exp_state in ucommitted_tests:
            with self.subTest(name=name):
                mock_api.journal_state.return_value = {
                    "uncommitted_journal_entries": ucommitted
                }
                state, _ = check_graylog_health.journal_state(mock_api, args)
                self.assertEqual(state, exp_state)


etc_default_conf = u"""
# Path to the java executable.
JAVA=/usr/bin/java

# Default Java options for heap and garbage collection.
{}

# Pass some extra args to graylog-server. (i.e. "-d" to enable debug mode)
GRAYLOG_SERVER_ARGS=""

# Program that will be used to wrap the graylog-server command. Useful to
# support programs like authbind.
GRAYLOG_COMMAND_WRAPPER=""
"""

heap_default = u"""
GRAYLOG_SERVER_JAVA_OPTS="-Xms1g -Xmx1g -XX:NewRatio=1 -server -XX:+ResizeTLAB \
-XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled \
-XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow"
"""

heap_2g = u"""
GRAYLOG_SERVER_JAVA_OPTS="-Xms2g -Xmx2g -XX:NewRatio=1 -server -XX:+ResizeTLAB \
-XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled \
-XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow"
"""

heap_not_present = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB \
-XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled \
-XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow -Xms4g -Xmx4g"
"""

heap_unset = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB \
-XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled \
-XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow"
"""

heap_not_present = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB \
-XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled \
-XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow -Xms4g -Xmx4g"
"""

heap_twice = u"""
GRAYLOG_SERVER_JAVA_OPTS="-XX:NewRatio=1 -server -XX:+ResizeTLAB \
-XX:+UseConcMarkSweepGC -XX:+CMSConcurrentMTEnabled -XX:+CMSClassUnloadingEnabled \
-XX:+UseParNewGC -XX:-OmitStackTraceInFastThrow -Xms8g -Xmx8g"
"""


class TestSetJVMHeapSize(unittest.TestCase):
    """Test JVM Heap size updates."""

    def setUp(self):  # noqa: D102
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        temp_file.write(etc_default_conf.format(heap_default).encode())
        temp_file.close()
        self.conf_file = temp_file.name

    def tearDown(self):  # noqa: D102
        os.remove(self.conf_file)

    def conf_equals(self, want):  # noqa: D102
        with open(self.conf_file, "r") as conf:
            got = conf.read()

            if got == want:
                return True
            print("{}\n--- != ---\n{}".format(got, want))

        return False

    def test_same(self):  # noqa: D102
        set_jvm_heap_size("1g", self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_default)))

    def test_update(self):  # noqa: D102
        set_jvm_heap_size("2g", self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_2g)))

    def test_unset(self):  # noqa: D102
        # Remove Xms and Xmx
        set_jvm_heap_size("", self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_unset)))

    def test_add(self):  # noqa: D102
        set_jvm_heap_size("", self.conf_file)
        set_jvm_heap_size("4g", self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_not_present)))

    def test_add_twice(self):  # noqa: D102
        set_jvm_heap_size("", self.conf_file)
        set_jvm_heap_size("6g", self.conf_file)
        set_jvm_heap_size("8g", self.conf_file)
        self.assertTrue(self.conf_equals(etc_default_conf.format(heap_twice)))

    def test_get_total_mem(self):
        """Test get_total_mem is working."""
        self.assertGreater(get_total_mem(), 0)

    def test_validate_jvm_heap_size_too_small(self):
        """Test jvm heap size too small or missing unit."""
        with self.assertRaises(ValueError) as context_manager:
            validate_jvm_heap_size("4")
        exc = context_manager.exception
        self.assertIn("must be greater than 2M", str(exc))

    def test_validate_jvm_heap_size_multiple_1024(self):
        """Test jvm heap size invalid bytes."""
        with self.assertRaises(ValueError) as context_manager:
            validate_jvm_heap_size("2049")
        exc = context_manager.exception
        self.assertIn("must be a multiple of 1024", str(exc))

    def test_validate_jvm_heap_size_invalid_format(self):
        """Test jvm heap size invalid unit."""
        with self.assertRaises(ValueError) as context_manager:
            validate_jvm_heap_size("4Gi")
        exc = context_manager.exception
        self.assertIn("invalid format", str(exc))

    def test_validate_jvm_heap_size_too_large(self):
        """Test jvm heap size larger than machine total mem."""
        with self.assertRaises(ValueError) as context_manager:
            validate_jvm_heap_size("4G", total_mem=3 * 1024 * 1024 * 1024)  # 3G
        exc = context_manager.exception
        self.assertIn("must be smaller than machine total mem", str(exc))

    def test_validate_jvm_heap_size_valid(self):
        """Test valid jvm heap size should not raise exception."""
        try:
            validate_jvm_heap_size("2G", total_mem=3 * 1024 * 1024 * 1024)  # 3G
        except ValueError:
            self.fail("Unexpected ValueError raised")


class TestIsChannelValid(unittest.TestCase):
    """Verify snap channel management."""

    def test_happy_path(self):  # noqa: D102
        for track in ("2", "3"):
            for risk in ("stable", "candidate", "beta", "edge"):
                channel = "{}/{}".format(track, risk)
                self.assertTrue(_is_channel_valid(channel))

    def test_sad_path(self):  # noqa: D102
        for channel in (
            "fake/stable",  # invalid track
            "2",  # track w/o risk
        ):
            self.assertFalse(_is_channel_valid(channel))


class TestTlsClient(unittest.TestCase):
    """Test TLS Client integration."""

    @mock.patch("charmhelpers.core.host.mkdir")
    @mock.patch("charmhelpers.core.host.write_file")
    @mock.patch("charmhelpers.core.hookenv.charm_dir")
    @mock.patch("charms.layer.graylog.utils.is_v2")
    @mock.patch("reactive.graylog.set_conf")
    @mock.patch("reactive.graylog.hookenv.unit_public_ip")
    @mock.patch("reactive.graylog.hookenv.unit_private_ip")
    @mock.patch("reactive.graylog.install")
    @mock.patch("subprocess.check_call")
    def test_request_certificate(
        self,
        mock_check_call,
        mock_install,
        mock_private_ip,
        mock_public_ip,
        mock_set_conf,
        mock_is_v2,
        mock_charm_dir,
        mock_write_file,
        mock_mkdir,
    ):
        charm_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../")
        mock_charm_dir.return_value = charm_dir
        mock_is_v2.return_value = False
        mock_public_ip.return_value = "1.2.3.4"
        mock_private_ip.return_value = "5.6.7.8"
        mock_tls = mock.Mock()
        tls_request_certificate(mock_tls)

        tls_client_mock.request_server_cert.assert_called_with(
            "1.2.3.4",
            ["1.2.3.4", "5.6.7.8", socket.gethostname(), "127.0.0.1", "localhost"],
            crt_path=CERT_PATH,
            key_path=CERT_KEY_PATH,
        )
        mock_set_conf.assert_has_calls(
            [
                mock.call("rest_enable_tls", "true"),
                mock.call("rest_tls_cert_file", CERT_PATH),
                mock.call("rest_tls_key_file", CERT_KEY_PATH),
                mock.call("web_enable_tls", "true"),
                mock.call("web_tls_cert_file", CERT_PATH),
                mock.call("web_tls_key_file", CERT_KEY_PATH),
            ]
        )

        with open(os.path.join(charm_dir, "templates/sync-graylog-snap"), "rb") as f:
            sync_script = f.read().strip()

        with open(
            os.path.join(charm_dir, "templates/default-graylog-server"), "rb"
        ) as f:
            default_graylog_server = f.read().strip()

        mock_write_file.assert_has_calls(
            [
                mock.call(
                    "/etc/ca-certificates/update.d/sync-graylog-snap",
                    sync_script,
                    "root",
                    "root",
                    0o755,
                ),
                mock.call(
                    "/var/snap/graylog/current/default-graylog-server",
                    default_graylog_server,
                    "root",
                    "root",
                    0o644,
                ),
            ]
        )
        mock_check_call.assert_called_with(["update-ca-certificates"])


class TestEmailConfig(unittest.TestCase):
    """Email config."""

    @mock.patch("reactive.graylog.report_status")
    @mock.patch("reactive.graylog.remove_state")
    @mock.patch("reactive.graylog.set_state")
    @mock.patch("reactive.graylog.set_jvm_heap_size")
    @mock.patch("reactive.graylog.hookenv.application_version_set")
    @mock.patch("reactive.graylog.hookenv.open_port")
    @mock.patch("reactive.graylog.is_master")
    @mock.patch("reactive.graylog.hookenv.config")
    @mock.patch("reactive.graylog.os")
    @mock.patch("reactive.graylog.unitdata.kv")
    @mock.patch("reactive.graylog.get_api_protocol")
    @mock.patch("reactive.graylog.set_conf")
    @mock.patch("charms.layer.graylog.utils.is_v2")
    @mock.patch("reactive.graylog.is_v2")
    @mock.patch("charms.layer.graylog.snap")
    @mock.patch("charms.layer.graylog.utils.get_api_port")
    def test_00_default_config(
        self,
        mock_get_api_port,
        mock_snap,
        mock_v2,
        mock_v2_2,
        mock_set_conf,
        mock_api_protocol,
        mock_unitdata_kv,
        mock_os,
        mock_hookenv_config,
        mock_is_master,
        mock_hookenv_open_port,
        mock_application_version_set,
        mock_jvm_heap_size,
        mock_remove_state,
        mock_set_state,
        mock_report_status,
    ):
        mock_hookenv_config.return_value = Config(
            {
                "jvm_heap_size": "1G",
                "rest_transport_uri": "foo/api",
                "web_endpoint_uri": "bar/api",
                "email_enabled": False,
                "email_hostname": "mail.example.com",
                "email_port": 587,
                "email_use_auth": False,
                "email_tls_enabled": False,
                "email_ssl_enabled": False,
                "email_auth_username": "you@example.com",
                "email_auth_password": "secret",
                "email_subject_prefix": "[graylog]",
                "email_from": "graylog@example.com",
                "email_web_url": "https://graylog.example.com",
            }
        )
        mock_unitdata_kv.return_value = {
            "admin_password": "password",
            "password_secret": "secret",
        }

        mock_v2.return_value = mock_v2_2.return_value = False
        mock_os.path.isfile.return_value = True
        mock_os.path.getsize.return_value = 1
        mock_get_api_port.return_value = None
        mock_api_protocol.return_value = "http"
        mock_snap.get_installed_version.return_value = "2.foo"
        mock_is_master.return_value = True

        configure_graylog()

        self.assertFalse(mock_set_conf.email_enabled.called)
        mock_set_conf.email_hostname.assert_not_called()

    @mock.patch("reactive.graylog.report_status")
    @mock.patch("reactive.graylog.remove_state")
    @mock.patch("reactive.graylog.set_state")
    @mock.patch("reactive.graylog.set_jvm_heap_size")
    @mock.patch("reactive.graylog.hookenv.application_version_set")
    @mock.patch("reactive.graylog.hookenv.open_port")
    @mock.patch("reactive.graylog.is_master")
    @mock.patch("reactive.graylog.hookenv.config")
    @mock.patch("reactive.graylog.os")
    @mock.patch("reactive.graylog.unitdata.kv")
    @mock.patch("reactive.graylog.get_api_protocol")
    @mock.patch("reactive.graylog.set_conf")
    @mock.patch("charms.layer.graylog.utils.is_v2")
    @mock.patch("reactive.graylog.is_v2")
    @mock.patch("charms.layer.graylog.snap")
    @mock.patch("charms.layer.graylog.utils.get_api_port")
    def test_01_modified_config(
        self,
        mock_get_api_port,
        mock_snap,
        mock_v2,
        mock_v2_2,
        mock_set_conf,
        mock_api_protocol,
        mock_unitdata_kv,
        mock_os,
        mock_hookenv_config,
        mock_is_master,
        mock_hookenv_open_port,
        mock_application_version_set,
        mock_jvm_heap_size,
        mock_remove_state,
        mock_set_state,
        mock_report_status,
    ):
        mock_hookenv_config.return_value = Config(
            {
                "jvm_heap_size": "1G",
                "rest_transport_uri": "foo/api",
                "web_endpoint_uri": "bar/api",
                "email_enabled": True,
                "email_hostname": "localhost",
                "email_port": 25,
                "email_use_auth": True,
                "email_tls_enabled": True,
                "email_ssl_enabled": True,
                "email_auth_username": "admin",
                "email_auth_password": "graylog",
                "email_subject_prefix": "[graylog_test]",
                "email_from": "graylog@localhost.com",
                "email_web_url": "http://graylog.localhost.com",
            }
        )
        mock_unitdata_kv.return_value = {
            "admin_password": "password",
            "password_secret": "secret",
        }

        mock_v2.return_value = mock_v2_2.return_value = False
        mock_os.path.isfile.return_value = True
        mock_os.path.getsize.return_value = 1
        mock_get_api_port.return_value = None
        mock_api_protocol.return_value = "http"
        mock_snap.get_installed_version.return_value = "2.foo"
        mock_is_master.return_value = True

        configure_graylog()

        mock_set_conf.assert_any_call("transport_email_enabled", "True")
        mock_set_conf.assert_any_call("transport_email_hostname", "localhost")
        mock_set_conf.assert_any_call("transport_email_port", "25")
        mock_set_conf.assert_any_call("transport_email_use_auth", "True")
        mock_set_conf.assert_any_call("transport_email_use_tls", "True")
        mock_set_conf.assert_any_call("transport_email_use_ssl", "True")
        mock_set_conf.assert_any_call("transport_email_auth_username", "admin")
        mock_set_conf.assert_any_call("transport_email_auth_password", "graylog")
        mock_set_conf.assert_any_call(
            "transport_email_subject_prefix", "[graylog_test]"
        )
        mock_set_conf.assert_any_call(
            "transport_email_from_email", "graylog@localhost.com"
        )
        mock_set_conf.assert_any_call(
            "transport_email_web_interface_url", "http://graylog.localhost.com"
        )

    @mock.patch("charmhelpers.core.hookenv.config")
    @mock.patch("charmhelpers.core.hookenv.is_leader")
    @mock.patch("charmhelpers.core.hookenv.relation_set")
    @mock.patch("charmhelpers.core.hookenv.unit_public_ip")
    @mock.patch("urllib.parse.urlparse")
    @mock.patch("charmhelpers.core.hookenv.relation_ids")
    @mock.patch("os.environ.get")
    def test_application_dashboard_joined(
        self,
        mock_environ,
        mock_relation_ids,
        mock_urlparse,
        mock_unit_public_ip,
        mock_relation_set,
        mock_is_leader,
        mock_hookenv_config,
    ):
        """Test registration relation joined."""
        mock_environ.return_value = ""
        mock_is_leader.return_value = True
        relation_id = "application-dashboard:0"
        mock_relation_ids.return_value = [relation_id]
        mock_unit_public_ip.return_value = "1.2.3.4"
        site_name = "test_site"
        mock_hookenv_config.return_value = {
            "web_listen_uri": "http://graylog-test:9001",
            "site_name": site_name,
        }
        application_dashboard_relation_changed("application-dashboard:0")
        name = "Graylog"
        subtitle = "[{}] Log management".format(site_name)
        group = "[{}] LMA".format(site_name)
        mock_relation_set.assert_called_once_with(
            relation_id,
            app=True,
            relation_settings={
                "name": name,
                "url": "http://1.2.3.4:9001",
                "subtitle": subtitle,
                "icon": None,
                "group": group,
            },
        )


class Config(dict):
    def __init__(self, *args, **kw):
        super(Config, self).__init__(*args, **kw)
